----------------------------------------------------------------
--
----------------------------------------------------------------

local addon = KiwiPlates

local Media  = LibStub("LibSharedMedia-3.0", true)

local pairs   = pairs
local ipairs  = ipairs
local strsub  = strsub
local format  = string.format
local tinsert = table.insert
local tremove = table.remove
local tconcat = table.concat

local UnitGUID = UnitGUID
local UnitName = UnitName
local UnitLevel = UnitLevel
local UnitIsUnit = UnitIsUnit
local UnitReaction = UnitReaction
local UnitClassification = UnitClassification
local C_GetNamePlateForUnit  = C_NamePlate.GetNamePlateForUnit
local C_SetNamePlateSelfSize = C_NamePlate.SetNamePlateSelfSize

local targetFrame
local GetPlateSkin
local ConditionFields = {}
	
local Reactions = { 'hostile', 'hostile', 'hostile', 'neutral', 'friendly', 'friendly',	'friendly',	'friendly' }
local Types = { a = 'Player', e = 'Creature', t = 'Pet', m = 'GameObject', h = 'Vehicle', g = 'Vignette' }

----------------------------------------------------------------
-- 
----------------------------------------------------------------

addon.defaults = {
	general     = {},
	skins       = { { __skinName = 'Default' } },
	rules       = { { 'and'  } },
}

----------------------------------------------------------------
-- Register media stuff early
----------------------------------------------------------------

Media:Register('font', 'Yanone Kaffesatz Bold', "Interface\\Addons\\KiwiPlates\\media\\yanone.ttf" )
Media:Register('font', 'FrancoisOne', "Interface\\Addons\\KiwiPlates\\media\\francois.ttf" )
Media:Register("statusbar", "Gradient", "Interface\\Addons\\KiwiPlates\\media\\gradient")
Media:Register("statusbar", "Blizzard Solid White", "Interface\\Buttons\\white8x8")
Media:Register("statusbar", "Blizzard NamePlate", "Interface\\TargetingFrame\\UI-TargetingFrame-BarFill")

----------------------------------------------------------------
-- Blizzard ShouldShowName() global function override
----------------------------------------------------------------

local ShouldShowName_Orig = ShouldShowName
local function ShouldShowName_Override()
	return true
end

----------------------------------------------------------------
-- CastBar visibility control (Interface\FrameXML\CastingBarFrame.lua)
----------------------------------------------------------------
	
do	
	local function CastBar_Event(self)
		if self.__hidden then 
			self:Hide() 
		end	
	end
	hooksecurefunc( "CastingBarFrame_OnShow", CastBar_Event )
	hooksecurefunc( "CastingBarFrame_OnUpdate", CastBar_Event )
end

----------------------------------------------------------------
-- Overrided method (Addons\Blizzard_Nameplates\Blizzard_NamePlates.lua)
----------------------------------------------------------------

local OnTargetChanged_Orig = NamePlateDriverFrame.OnTargetChanged

local function OnTargetChanged(self)
	if targetFrame then
		local UnitFrame = targetFrame.UnitFrame
		UnitFrame.__target = nil
		addon:SkinPlate(targetFrame, UnitFrame)
		targetFrame = nil
	end
	local plateFrame = C_GetNamePlateForUnit('target')
	if plateFrame then
		local UnitFrame = plateFrame.UnitFrame
		UnitFrame.__target = true
		addon:SkinPlate(plateFrame, UnitFrame)
		targetFrame = plateFrame
	end	
	self:SetupClassNameplateBars()
	self:OnUnitAuraUpdate("target")
end

----------------------------------------------------------------
--
----------------------------------------------------------------
	
function addon:NAME_PLATE_UNIT_ADDED(unit)
	local plateFrame = C_GetNamePlateForUnit(unit)
	if plateFrame then
		local UnitFrame = plateFrame.UnitFrame
		if UnitIsUnit(unit, 'target') then
			targetFrame = plateFrame
			UnitFrame.__target = true
		end
		if not UnitFrame.__valid then
			UnitFrame.__valid  = true		
			UnitFrame.__target = UnitIsUnit( unit, 'target' ) or nil
			UnitFrame.__name   = UnitName( unit )	
			UnitFrame.__level  = UnitLevel( unit )		
			UnitFrame.__type   = Types[ strsub( UnitGUID(unit), 3,3 ) ]
			UnitFrame.__reaction = Reactions [ UnitReaction( unit, "player") or 1 ]		
			UnitFrame.__classification = UnitClassification(unit) or 'unknow'
			if UnitFrame.__level==-1 or UnitFrame.__classification=='worldboss' then	
				UnitFrame.__classification = 'boss' 
			end
			self:SkinPlate( plateFrame, UnitFrame)
		end
		return UnitFrame
	end	
end

----------------------------------------------------------------
--
----------------------------------------------------------------

function addon:NAME_PLATE_UNIT_REMOVED(unit)
	local plateFrame = C_GetNamePlateForUnit(unit)
	local UnitFrame = plateFrame.UnitFrame
	UnitFrame.__valid  = nil
	UnitFrame.__threat = nil
	UnitFrame.__target = nil
	if plateFrame == targetFrame then
		targetFrame = nil
	end	
end

----------------------------------------------------------------
--
----------------------------------------------------------------
	
function addon:SkinPlate(plateFrame, UnitFrame)

	local db = addon.db.skins[ GetPlateSkin(UnitFrame) ]
	
	local healthBar = UnitFrame.healthBar
	healthBar:SetHeight( db.healthBarHeight or 4 )	

	local castBar = UnitFrame.castBar	
	local cbHeight = db.castBarHeight or 8
	castBar:ClearAllPoints()
	castBar:SetPoint('BOTTOM', plateFrame, 'BOTTOM', 0, 6)
	castBar:SetSize( db.healthBarWidth or 110, cbHeight )
	castBar:SetStatusBarTexture( Media:Fetch("statusbar", db.castBarTexture or "Blizzard NamePlate" ) )
	castBar.Icon:SetSize(cbHeight, cbHeight)
	castBar.__hidden = db.castBarHidden
	
	if UnitFrame.__classification=='boss' then
		local classificationIndicator = UnitFrame.ClassificationFrame.classificationIndicator
		classificationIndicator:SetTexture("Interface\\TargetingFrame\\UI-TargetingFrame-Skull")
		classificationIndicator:SetTexCoord(0.1, 0.9, 0, 1)
	end

	if db == UnitFrame.__skin then return end 

	healthBar:SetShown(not db.healthBarHidden)
	
	castBar.Icon:SetTexCoord( 0.1, 0.9, 0.1, 0.9 )
	castBar.BorderShield:SetSize(cbHeight, cbHeight)	
	castBar.Text:SetFont( Media:Fetch('font', db.castBarFontFile or 'Friz Quadrata TT') or "Fonts\\FRIZQT__.TTF" , db.castBarFontSize or 8, db.castBarFontFlags or '' )

	healthBar:SetStatusBarTexture( Media:Fetch("statusbar", db.healthBarTexture or "Blizzard NamePlate" ) )
	
	local RaidTargetFrame = UnitFrame.RaidTargetFrame
	RaidTargetFrame:SetPoint("RIGHT", healthBar, "LEFT", db.raidTargetOffsetX or -15, db.raidTargetOffsetY or 0);
	RaidTargetFrame:SetSize( db.raidTargetSize or 22, db.raidTargetSize or 22 )

	local name = UnitFrame.name
	name:SetPoint("BOTTOM", healthBar, "TOP", db.nameOffsetX or 0, db.nameOffsetY or 4);
	name:SetFont( Media:Fetch('font', db.nameFontFile or 'Friz Quadrata TT') or "Fonts\\FRIZQT__.TTF" , db.nameFontSize or 11, db.nameFontFlags or '' )

	local ClassificationFrame = UnitFrame.ClassificationFrame
	local size = db.classIconSize or 14
	ClassificationFrame:SetSize( size, size )
	ClassificationFrame:SetPoint('RIGHT', healthBar, 'LEFT', db.classIconOffsetX or 0, db.classIconOffsetY or 0)	
	ClassificationFrame:Show()

	if addon.db.auras.enabled then
		UnitFrame.BuffFrame:SetPoint("BOTTOM", healthBar, "TOP", db.buffsOffsetX or 0, db.buffsOffsetY or 0)
	end	
	
	UnitFrame.__skin = db		
	
end	

----------------------------------------------------------------
-- Compile a function to calculate nameplate skin
----------------------------------------------------------------

do
	local function NamesToList(names)
		local lines = {}
		local t = { strsplit("\n",names) }
		for i=1,#t do -- Remove comments: any text starting with #@\/-[ characters.
			local s = strtrim( (strsplit( "#@\\\/\-\[", t[i] )) ) -- Don't remove strsplit extra brackets.
			if #s>0 then tinsert( lines, format('["%s"]=true',s) ) end
		end
		return tconcat( lines , ',')
	end
	local function MakeSkinCheckWithClosure(names, source)
		local lines = { "return function()" }
		for i=1,#names do
			tinsert( lines, format("local names%d = {%s}",i, NamesToList(names[i]) ) )
		end
		tinsert( lines,  source )
		tinsert( lines, 'end' )
		return assert(loadstring(tconcat( lines, "\n")))()()
	end
	function addon:CompileSkinCheckFunction(rules, fields)
		if fields then wipe(fields) end
		local count, handler, lines, names, iff = 0, { 'return function(p)' }, {}, {}, 'if'
		for j=1,#rules do
			local rule = rules[j]
			if #rule>1 then 
				for i=2,#rule do
					local c = rule[i]
					if c[1] == 'names' then
						tinsert( names, c[3] )
						tinsert( lines, format("names%d[p.__name]",#names) )
					else
						local typ, line = type(c[3])
						if c[3] == 'nil' then
							line = format( 'p.__%s == nil', c[1] )
						elseif typ == 'string' then
							line = format( 'p.__%s %s "%s"', c[1], c[2], c[3] )
						elseif typ == 'boolean' then
							line = format( 'not p.__%s %s %s', c[1], c[2], tostring(not c[3]) )
						else
							line = format( 'p.__%s %s %s', c[1], c[2], tostring(c[3]) )
						end
						tinsert( lines, line )
					end
					if fields then fields[c[1]] = true end
				end
				tinsert( handler, format( '%s %s then return %d', iff, tconcat(lines,format(' %s ',rule[1])), j) )
				iff = "elseif"
				wipe(lines)
				count = count + 1
			end	
		end
		tinsert( handler, count>0 and 'else return 1 end end' or 'return 1 end' )
		if #names>0 then
			return MakeSkinCheckWithClosure( names, tconcat(handler,"\n") )
		else	
			return assert(loadstring(tconcat(handler,"\n")))()
		end
	end
end

----------------------------------------------------------------
-- 
----------------------------------------------------------------

function addon:OnProfileChanged()
	self:Update(true) 
end

----------------------------------------------------------------
--
----------------------------------------------------------------

function addon:Update(forceReskin)

	_G['ShouldShowName'] = self.db.general.namesAlwaysVisible and ShouldShowName_Override or ShouldShowName_Orig
	
	local flag = (#self.db.skins>1 or next(self.db.skins[1],next(self.db.skins[1]))~=nil) or nil
	if flag ~= self.__userDefinedPlates then
		if flag then
			self:RegisterEvent("NAME_PLATE_UNIT_ADDED")
			self:RegisterEvent("NAME_PLATE_UNIT_REMOVED")
		else
			self:UnregisterEvent("NAME_PLATE_UNIT_ADDED")
			self:UnregisterEvent("NAME_PLATE_UNIT_REMOVED")
		end
		self.__userDefinedPlates = flag
	end	
	
	GetPlateSkin = self:CompileSkinCheckFunction(addon.db.rules, ConditionFields)

	NamePlateDriverFrame.OnTargetChanged = ConditionFields.target and OnTargetChanged or OnTargetChanged_Orig
	
	for _, module in pairs(self.modules) do
		local func = module.Update
		if func then func(self, forceReskin) end
	end
	
	for _, plateFrame in pairs(C_NamePlate.GetNamePlates()) do
		local UnitFrame = plateFrame.UnitFrame
		if forceReskin then UnitFrame.__skin = nil end
		self:SkinPlate(plateFrame, UnitFrame)
	end	
	
end
